(* Wilson's algorithm as described in
 * https://en.wikipedia.org/wiki/Maze_generation_algorithm
 * at 2024-05-22
 *)

let rows = 25
let cols = 25

let get arr (col, row) = arr.(col * cols + row)
let set arr (col, row) v = arr.(col * cols + row) <- v
         
let is_on_grid arr (col, row) =
  not (row < 0 || row >= rows || col < 0 || col >= cols)

let is_cell arr (row, col) =
  if not @@ is_on_grid arr (row, col) then false
  else get arr (row, col) <> ' '

let ( ++ ) (a, b) (x, y) = (a + x, b + y)
let ( // ) (a, b) n = (a / n, b / n)
                
let all_deltas = [(0, 2); (0, -2); (2, 0); (-2, 0)]

let all_neighbors arr pos =
  List.filter (fun v -> is_on_grid arr v)
    (List.map (fun v -> v ++ pos) all_deltas)

let random_element ls = List.nth ls (Random.int (List.length ls))

let list_find_tail haystack needle =
  let rec loop = function
    | [] -> haystack
    | x :: xs when x = needle -> xs
    | _ :: xs -> loop xs in
  loop haystack

let rec insert_conn path acc = match path with
  | [] -> assert false
  | [p] -> p :: acc
  | p :: ((p0 :: _) as rest) -> insert_conn rest (((p ++ p0) // 2) :: p :: acc)
  
let rec random_walk arr pos path =
  let neighbors = all_neighbors arr pos in
  let neighbor = random_element neighbors in
  if get arr neighbor <> ' ' then
    List.rev (neighbor :: path)
  else
    let path = neighbor :: list_find_tail path neighbor in
    random_walk arr neighbor path

let random_empty_cell arr =
  let empty_cells = ref [] in
  for row = 0 to rows - 1 do
    for col = 0 to cols - 1 do
      if (row mod 2 = 0) && (col mod 2 = 0) && get arr (row, col) = ' ' then
        empty_cells := (row, col) :: !empty_cells
    done;
  done;
  Printf.printf "empty cells: %d\n" (List.length !empty_cells);
  match !empty_cells with
  | [] -> None
  | ls -> Some (random_element ls)

let output_maze stream arr = 
  for row = 0 to rows - 1 do
    for col = 0 to cols - 1 do
      output_char stream (get arr (row, col))
    done;
    output_char stream '\n';
  done

let rec draw_maze arr =
  match random_empty_cell arr with
  | None -> ()
  | Some pos ->
     let path = pos :: random_walk arr pos [] in
     let path = insert_conn path [] in
     Printf.printf "path: ";
     List.iter (fun (r, c) -> Printf.printf "(%d,%d) " r c) path;
     Printf.printf "\n";
     List.iter (fun pos -> set arr pos '+') path;
     draw_maze arr
  
let place_eggs_boosters arr =
  for row = 0 to rows - 1 do
    for col = 0 to cols - 1 do
      if is_cell arr (row, col) then
        let f delta char =
          if not (is_cell arr ((row, col) ++ delta))
             && Random.int 256 < 10 then
            set arr (row, col) char in
        begin
          f (1, 0) 'v';
          f (-1, 0) '^';
          f (0, 1) '>';
          f (0, -1) '<';
          if Random.int 512 < 4 then
            set arr (row, col) 'B';
        end
    done
  done
  
let main () =
  Random.init 4;
  let arr = Array.make (rows * cols) ' ' in
  arr.(0) <- '+';
  draw_maze arr;
  place_eggs_boosters arr;
  set arr (0, 0) 'R';
  set arr (0, cols - 1) '^';
  set arr (rows - 1, 0) 'v';
  set arr (rows - 1, cols - 1) 'O';
  let stream = open_out "voxelmaze7.vox" in
  output_string stream "#floor_tex shared_pk02/floor01 0.25 0.25 0 0
#wall_tex_random shared_pk02/wall02a_norad 0.5 0.5 0 0
#wall_tex_random shared_pk02/wall02b_norad 0.5 0.5 0 0
#wall_tex_random shared_pk02/wall02c_norad 0.5 0.5 0 0
#sky_tex voxel/sky 1 1 0 0
#single_sky
#eggs_dist 86
#eggs_num 2
";
  output_maze stream arr;
  close_out stream

let () = main ()
